from unittest.mock import MagicMock

from django import forms
from django.conf import settings
from django.contrib.admin.widgets import SELECT2_TRANSLATIONS
from django.contrib.admin.widgets import (
    AutocompleteSelect as BaseAutocompleteSelect,
)
from django.contrib.admin.widgets import (
    AutocompleteSelectMultiple as BaseAutocompleteSelectMultiple,
)
from django.db import models
from django.utils.translation import get_language

from jnt_django_toolbox.admin.helpers.widgets import render_autocomplete_badge


class AdminAutocompleteError(Exception):
    def __init__(self, model):
        message = "Related objects for '{0}' not found. Can't create autocomplete widget.".format(
            model._meta.object_name,
        )
        super().__init__(message)


class AutocompleteMixin:
    """Autocomplete mixin."""

    def __init__(self, field, *args, **kwargs):
        """Init autocomplete mixin."""
        if isinstance(field, type) and issubclass(field, models.Model):
            for related_object in field._meta.related_objects:
                if related_object.model is field:
                    field = related_object.remote_field
                    break
            else:
                field = _fake_field(field)

        self._query_params = kwargs.pop("query_params", {})
        super().__init__(field, *args, **kwargs)

    @property
    def media(self):
        """Media."""
        extra = "" if settings.DEBUG else ".min"
        i18n_name = SELECT2_TRANSLATIONS.get(get_language())
        i18n_file = (
            ("admin/js/vendor/select2/i18n/{0}.js".format(i18n_name),)
            if i18n_name
            else ()
        )
        return forms.Media(
            js=(
                "admin/js/vendor/jquery/jquery{0}.js".format(extra),
                "admin/js/vendor/select2/select2.full{0}.js".format(extra),
            )
            + i18n_file
            + (
                "admin/js/jquery.init.js",
                "jnt_django_toolbox/js/widgets/autocomplete.js",
            ),
            css={
                "screen": (
                    "admin/css/vendor/select2/select2{0}.css".format(extra),
                    "admin/css/autocomplete.css",
                ),
            },
        )

    def build_attrs(self, base_attrs, extra_attrs=None):
        widget_attrs = super().build_attrs(base_attrs, extra_attrs=extra_attrs)
        for attr, attr_value in self._query_params.items():
            widget_attrs["data-autocomplete--{0}".format(attr)] = attr_value
        return widget_attrs


class AutocompleteSelect(AutocompleteMixin, BaseAutocompleteSelect):
    """Autocomplete select widget."""


class AutocompleteSelectMultiple(
    AutocompleteMixin, BaseAutocompleteSelectMultiple
):
    """Autocomplete select multiple widget."""

    def optgroups(self, name, value, attr=None):
        """Return a list of optgroups for this widget."""
        original_label_from_instance = self.choices.field.label_from_instance

        self.choices.field.label_from_instance = (
            self._autocomplete_label_from_instance
        )

        option_groups = super().optgroups(name, value, attr=attr)

        self.choices.field.label_from_instance = original_label_from_instance

        return option_groups

    def _autocomplete_label_from_instance(self, instance, present=None):
        return render_autocomplete_badge(instance, present, self.admin_site)


def _fake_field(model: models.Model):
    """Fake field factory."""
    mock_field = MagicMock()
    mock_field.model = model
    mock_field.name = ""
    mock_field.remote_field.model = model
    mock_field.remote_field.field_name = "id"
    return mock_field



from functools import wraps

from jnt_django_toolbox.context_managers import global_lock


def one_at_time(func):
    """Ensures that function runned only one at time."""

    @wraps(func)
    def handle(*args, **kwargs):  # noqa:  WPS110
        with global_lock(func.__name__) as acquired:
            if acquired:
                return func(*args, **kwargs)

    return handle



import math
import re

ELLIPSIS = "..."
MAX_SUFFIX = 3


def truncate_text(text: str, max_length: int = 50) -> str:
    """Truncate long text."""
    if not text or not max_length:
        return ""

    len_ellipsis = len(ELLIPSIS)

    text = text.replace("\r", "").replace("\n", "")

    if max_length >= len(text):
        return text
    elif len_ellipsis >= max_length:
        return text[:max_length]

    start = max_length - len_ellipsis - MAX_SUFFIX

    if start < MAX_SUFFIX:
        start = math.ceil((max_length - len_ellipsis) / 2)

    end = max_length - start - len_ellipsis

    return "{0}{1}{2}".format(
        text[:start],
        ELLIPSIS,
        text[-end:] if end > 0 else "",
    )


def camel_case_to_snack_case(text: str) -> str:
    """Convert CamelCase -> snake_case."""
    text = _validate_text(text)

    text = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", text).lower()


def snack_case_to_camel_case(text: str) -> str:
    """Convert snake_case -> CamelCase."""
    text = _validate_text(text)

    first, *words = text.split("_")
    return "{0}{1}".format(first, "".join(word.title() for word in words))


def remove_prefix(text: str, prefix: str) -> str:
    """
    Remove prefix for string.

    Obsolete with python3.9+
    """
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def _validate_text(text: str) -> str:
    """Text validation for convert snack <-> camel cases."""
    text = text.strip()

    if re.search(r"\s", text):
        raise ValueError("The text should not contain spaces")

    return text



def ensure_fields_saved(save_kwargs, fields):
    """Ensuring fields saving (extending "update_fields")."""
    update_fields = save_kwargs.get("update_fields")
    if not update_fields:
        return

    save_kwargs["update_fields"] = tuple(set(list(update_fields) + fields))



import abc


class BaseReadOnlyWidget(abc.ABC):
    """Base readonly widget."""

    @abc.abstractmethod
    def render(self, field_value, field_name, **kwargs) -> str:
        """Render current widget."""



from typing import Optional

from django.contrib.admin.utils import lookup_field
from django.core.exceptions import FieldDoesNotExist, ObjectDoesNotExist


def get_present_admin_readonly_field(  # noqa: WPS212
    admin_readonly_field,
) -> Optional[str]:
    """Get present for AdminReadonlyField."""
    field_name, instance, model_admin = (
        admin_readonly_field.field["field"],
        admin_readonly_field.form.instance,
        admin_readonly_field.model_admin,
    )

    if not instance or admin_readonly_field.is_checkbox:
        return None

    try:
        field, attr, field_value = lookup_field(
            field_name,
            instance,
            model_admin,
        )
        if field is None:
            field = instance._meta.get_field(field_name)
    except (AttributeError, ValueError, ObjectDoesNotExist, FieldDoesNotExist):
        # custom values can be present at "form.initial"
        field_name = admin_readonly_field.field["name"]
        field_value = admin_readonly_field.form.initial.get(field_name)
        field = None
        if all((not field_value, not isinstance(field_value, bool))):
            return None

    func_readonly_widget = getattr(model_admin, "readonly_widget", None)
    if not func_readonly_widget:
        return None

    readonly_widget = func_readonly_widget(admin_readonly_field.field["field"])

    if readonly_widget:
        return readonly_widget.render(field_value, field_name, db_field=field)

    return None



from django.contrib.admin.templatetags.admin_list import (  # noqa: WPS450
    _boolean_icon as boolean_icon,
)

from jnt_django_toolbox.forms.widgets.readonly.base import BaseReadOnlyWidget


class BooleanReadonlyWidget(BaseReadOnlyWidget):
    def render(self, field_value, field_name, **kwargs) -> str:
        """Render boolean readonly widget."""
        return boolean_icon(field_value)



from django.forms.models import ModelChoiceIterator

from jnt_django_toolbox.forms.widgets import PermissionSelectMultipleWidget


class FakeField:
    """Fake field for ModelChoiceIterator."""

    def __init__(self, queryset) -> None:
        self.queryset = queryset
        self.empty_label = False


class PermissionSelectMultipleReadonlyWidget(PermissionSelectMultipleWidget):
    """Permissions multiple readonly widget."""

    def __init__(self, *args, **kwargs) -> None:
        """Initializing."""
        super().__init__(*args, **kwargs)
        self.extra_data["is_readonly"] = True

    def render(self, value, name, *args, **kwargs):
        """Render html-string."""
        permissions = value.all()

        self.choices = ModelChoiceIterator(FakeField(permissions))
        self.groups_permissions = permissions

        return super().render(name=name, value=permissions)



import typing as ty

from django.contrib import admin
from django.contrib.auth import get_permission_codename
from django.db import models, transaction

from jnt_django_toolbox.admin.helpers.urls import admin_url_provider
from jnt_django_toolbox.admin.views.base_admin_page import (
    AdminPageBreadcrumb,
    BaseAdminPageView,
)


class BaseModelAdminPageView(BaseAdminPageView):
    """Base admin view for object related pages."""

    model_admin: admin.ModelAdmin = None

    def get_model(self) -> models.Model:
        return self.model_admin.model

    def get_permission_required(self) -> ty.Iterable[str]:
        if self.permission_required:
            return super().get_permission_required()

        opts = self.get_model()._meta

        return (
            "{0}.{1}".format(
                opts.app_label,
                get_permission_codename("add", opts),
            ),
        )

    @transaction.atomic
    def post(self, *args, **kwargs):
        return super().post(*args, **kwargs)

    def get_breadcrumbs(self) -> list[AdminPageBreadcrumb]:
        model = self.get_model()
        breadcrumbs = [
            AdminPageBreadcrumb(
                title=model._meta.app_label,
                href="/{0}/{1}/".format(
                    model._meta.app_label,
                    self.model_admin.admin_site.name,
                ),
            ),
            AdminPageBreadcrumb(
                title=str(model._meta.verbose_name_plural)
                or model._meta.model_name,
                href=admin_url_provider.list_url(model),
            ),
        ]

        return [*breadcrumbs, *super().get_breadcrumbs()]



from django.contrib import admin
from django.urls import reverse


def render_autocomplete_badge(
    instance,
    present=None,
    admin_site: admin.AdminSite = None,
) -> str:
    admin_site = admin_site or admin.site

    return '<a href="{0}" target="_blank" title="{1}">{1}</a>'.format(
        reverse(
            "{0}:{1.app_label}_{1.model_name}_change".format(
                admin_site.name,
                instance._meta,
            ),
            args=(instance.id,),
        ),
        present or str(instance),
    )



from .base_admin_page import BaseAdminPageView, AdminPageBreadcrumb
from .base_instance_model_admin_page import BaseInstanceModelAdminPageView
from .base_model_admin_page import BaseModelAdminPageView



from django.forms import MultipleChoiceField, TypedChoiceField


class EnumChoiceField(TypedChoiceField):
    """Form field for enum."""


class MultipleEnumChoiceField(MultipleChoiceField):
    """Multiple form field for enum."""

    def __init__(self, enum, *args, **kwargs) -> None:
        """Init MultipleEnumChoiceField."""
        self.enum = enum
        kwargs["choices"] = enum.choices
        super().__init__(*args, **kwargs)



from django.core.management.commands import migrate

from jnt_django_toolbox.context_managers import global_lock

MIGRATE_LOCK_NAME = "db_migrate"


class Command(migrate.Command):
    """
    Migrate concurrently.

    Only 1 migrate will be runned at once.
    """

    def handle(self, *args, **options):  # noqa: WPS110
        """Command body."""
        with global_lock(MIGRATE_LOCK_NAME) as acquired:
            if acquired:
                super().handle(*args, **options)
            else:
                self.stdout.write(
                    self.style.WARNING("Migration already runned"),
                )



import typing as ty
from http import HTTPStatus

from django.contrib import admin
from django.contrib.admin.sites import site
from django.contrib.admin.views.autocomplete import AutocompleteJsonView
from django.http import Http404, JsonResponse
from django.urls import reverse
from django.utils.text import capfirst

from jnt_django_toolbox.admin.helpers.urls import admin_url_provider


class ContentTypeAutocompleteView(AutocompleteJsonView):
    # copied from base class
    def get(self, request, *args, **kwargs):
        if not self.model_admin.get_search_fields(request):
            msg_template = (
                "{0} must have search_fields for the autocomplete_view."
            )
            raise Http404(msg_template.format(type(self.model_admin).__name__))
        if not self.has_perm(request):
            return JsonResponse(
                {"error": "403 Forbidden"},
                status=HTTPStatus.FORBIDDEN,
            )

        self.term = request.GET.get("term", "")
        self.paginator_class = self.model_admin.paginator
        self.object_list = self.filter_queryset(self.get_queryset())
        context = self.get_context_data()

        return JsonResponse(
            {
                "results": [
                    {
                        "id": str(content_type.pk),
                        "text": capfirst(content_type.model),
                        # modify
                        "autocomplete_url": self.get_autocomplete_url(
                            content_type
                        ),
                        "data_app": content_type.app_label,
                        "data_model": content_type.model,
                        "data_change_url": self._get_change_url_template(
                            content_type
                        ),
                    }
                    for content_type in context["object_list"]
                ],
                "pagination": {"more": context["page_obj"].has_next()},
            }
        )

    def get_autocomplete_url(self, content_type):
        return admin_url_provider.autocomplete(content_type.model_class())

    def filter_queryset(self, queryset):
        return self._filter_queryset_by_ids(queryset)

    def _filter_queryset_by_ids(self, queryset):
        ids = self.request.GET.getlist("ids[]")
        if ids:
            queryset = queryset.filter(id__in=ids)
        return queryset

    def _get_change_url_template(self, content_type):
        list_url = reverse(
            "admin:{0}_{1}_changelist".format(
                content_type.app_label,
                content_type.model,
            ),
        )
        return "{0}{{id}}/change/".format(list_url)


def get_model_admin(model) -> ty.Optional[admin.ModelAdmin]:
    return site._registry.get(model)


class BaseContentTypeAdmin(admin.ModelAdmin):
    """Base content type admin."""

    search_fields = ("model",)

    def has_add_permission(self, request):
        """Check has add permissions."""
        return False

    def has_change_permission(self, request, obj=None):
        """Check has change permissions."""
        return False

    def has_view_or_change_permission(self, request, obj=None):
        """Check has view or change permissions."""
        return False

    def has_view_permission(self, request, obj=None):
        """Check has view permissions."""
        return True

    def get_model_perms(self, request):
        """Return a dict of all perms for this model."""
        return {}

    def autocomplete_view(self, request):
        """Content type autocomplete view."""
        return ContentTypeAutocompleteView.as_view(model_admin=self)(request)

    def autocomplete_queryset(self, request, queryset):
        queryset = super().autocomplete_queryset(request, queryset)

        ids = []
        for content_type in queryset:
            model_class = content_type.model_class()

            if get_model_admin(model_class):
                ids.append(content_type.pk)

        return queryset.filter(id__in=ids)

    def autocomplete_item_data(self, instance, request):
        item_data = super().autocomplete_item_data(instance, request)

        model = instance.model_class()

        related_obj = None

        for related_object in model._meta.related_objects:
            model_admin = get_model_admin(related_object.related_model)
            if model_admin:
                related_obj = related_object
                break

        if not related_obj:
            return item_data

        item_data.update(
            {
                "app_label": related_obj.related_model._meta.app_label,
                "model": related_obj.related_model._meta.model_name,
                "field_name": related_obj.remote_field.name,
                "field_placeholder": model._meta.verbose_name.title(),
            }
        )

        return item_data



from django.contrib import admin

from jnt_django_toolbox.forms.widgets.readonly.base import BaseReadOnlyWidget


class AdminChangeListField:
    """Admin changelist field."""

    def __init__(
        self,
        model_admin: admin.ModelAdmin,
        field_name: str,
        readonly_widget: BaseReadOnlyWidget,
        **kwargs,
    ) -> None:
        """Init column presenter."""
        self._model_admin = model_admin
        self._field_name = field_name
        self._readonly_widget = readonly_widget
        self._field_present = kwargs.get("field_present")
        self.__name__ = field_name

    def __call__(self, instance) -> str:
        """Implement callable for render widget."""
        return self._readonly_widget.render(
            getattr(instance, self._field_name),
            self._field_name,
            db_field=self._model_admin.model._meta.get_field(self._field_name),
            field_present=self._field_present,
        )



from django.template.response import TemplateResponse


class AutocompleteChangelistFiltersAdminMixin:
    def changelist_view(self, request, extra_context=None) -> TemplateResponse:
        changelist_view = super().changelist_view(
            request, extra_context=extra_context
        )
        self._update_changelist_media(changelist_view)
        return changelist_view

    def _update_changelist_media(
        self,
        changelist_view: TemplateResponse,
    ) -> None:
        try:
            context_data = changelist_view.context_data
        except AttributeError:
            return
        else:
            if "cl" not in context_data:
                return

        for filter_spec in changelist_view.context_data["cl"].filter_specs:
            if not getattr(filter_spec, "media", None):
                continue
            changelist_view.context_data["media"] += filter_spec.media



import collections

not_provided = object()


def recursive_dict():
    """Default dict of default dicts with unlimited nesting."""
    return collections.defaultdict(recursive_dict)


def deep_get(dic, path, default=not_provided):
    """Get a nested dict value by dot-separated keys."""
    keys = path.split(".")

    try:  # noqa:WPS229
        for key in keys[:-1]:
            dic = dic[key]

        return dic[keys[-1]]
    except KeyError:
        if default == not_provided:
            raise

        return default


def deep_set(dic, path, value_to_set):
    """Set a nested dict value by dot-separated keys."""
    keys = path.split(".")

    for key in keys[:-1]:
        dic = dic[key]

    dic[keys[-1]] = value_to_set


def set_value_if_none(data_dict, item_key: str, item_value: object) -> None:
    """Set value if dict value is null."""
    try:
        data_value = data_dict[item_key]
    except KeyError:
        return
    else:
        if data_value is None:
            data_dict[item_key] = item_value



from functools import partialmethod

from django import forms

from jnt_django_toolbox.admin.mixins import GenericForeignKeyAdminMixin


class GenericForeignKeyInlineAdminMixin(GenericForeignKeyAdminMixin):
    """Generic foreign key inline admin mixin."""

    def get_formset(self, request, obj=None, **kwargs):
        """Get formset."""
        formset = super().get_formset(request, obj, **kwargs)

        generic_relation_fields = self._get_generic_relation_fields(self.model)
        if not generic_relation_fields:
            return formset

        for generic_relation_field in generic_relation_fields:
            if generic_relation_field.fk_field in formset.form.base_fields:
                formset.form.base_fields[
                    generic_relation_field.fk_field
                ].widget = forms.HiddenInput()

                clean_func = partialmethod(
                    self._clean_fk_field,
                    fk_field=generic_relation_field.fk_field,
                    ct_field=generic_relation_field.ct_field,
                )
                setattr(
                    formset.form,
                    "clean_{0}".format(generic_relation_field.fk_field),
                    clean_func,
                )

        return formset



import typing

from jnt_django_toolbox.forms.widgets.readonly.base import BaseReadOnlyWidget


class StringReadonlyWidget(BaseReadOnlyWidget):
    def render(
        self,
        field_value,
        field_name,
        **kwargs,
    ) -> typing.Optional[str]:
        db_field = kwargs.get("db_field")
        if getattr(db_field, "choices", None):
            choice = [
                present
                for choice_value, present in db_field.choices
                if choice_value == field_value
            ]

            if choice:
                return choice[0]

        return field_value



from .permissions_readonly import PermissionSelectMultipleReadonlyWidget
from .foreign_key import ForeignKeyReadonlyWidget
from .many_to_many import ManyToManyReadonlyWidget
from .generic_foreign_key import GenericForeignKeyReadonlyWidget
from .string_field import StringReadonlyWidget
from .boolean import BooleanReadonlyWidget



import typing as ty

from django.contrib.auth import get_permission_codename
from django.db import models

from jnt_django_toolbox.admin.helpers.urls import admin_url_provider
from jnt_django_toolbox.admin.views import (
    AdminPageBreadcrumb,
    BaseModelAdminPageView,
)


class BaseInstanceModelAdminPageView(BaseModelAdminPageView):
    """Base admin view for object related pages."""

    instance: models.Model = None

    def get_permission_required(self) -> ty.Iterable[str]:
        if self.permission_required:
            return super().get_permission_required()

        opts = self.get_model()._meta

        return (
            "{0}.{1}".format(
                opts.app_label,
                get_permission_codename("change", opts),
            ),
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            instance=self.instance,
            subtitle=str(self.instance),
        )

        return context

    def get_breadcrumbs(self) -> list[AdminPageBreadcrumb]:
        breadcrumbs = list(super().get_breadcrumbs())
        breadcrumbs.insert(
            -1,
            AdminPageBreadcrumb(
                title=str(self.instance),
                href=admin_url_provider.change_url(
                    self.instance,
                    model_admin=self.model_admin,
                ),
            ),
        )

        return breadcrumbs



from django import template
from django.utils.translation import gettext

register = template.Library()


@register.filter
def get_item(d, key):
    return d.get(key, None)


@register.filter(name="translate")
def translate(text):
    return gettext(text)



from django import forms
from django.contrib.admin.widgets import (
    AutocompleteSelect as BaseAutocompleteSelect,
)
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse


class GenericForeignKeyWidget(BaseAutocompleteSelect):
    template_name = "jnt_django_toolbox/widgets/generic_foreign_key.html"

    @property
    def media(self):
        media = super().media

        media_js = media._js if media else []
        media_css = media._css if media else {}
        media_css["screen"] = list(media_css["screen"])
        media_css["screen"].append(
            "jnt_django_toolbox/css/widgets/generic-foreign-key-field.css",
        )

        return forms.Media(
            js=tuple(
                [
                    *media_js,
                    "jnt_django_toolbox/js/widgets/generic-foreign-key-field.js",
                ]
            ),
            css=media_css,
        )

    def create_option(
        self,
        name,
        value,
        label,
        selected,
        index,
        subindex=None,
        attrs=None,
    ):
        result = super().create_option(
            name,
            value,
            label,
            selected,
            index,
            subindex=subindex,
            attrs=attrs,
        )

        if value:
            self._update_attrs(result["attrs"], value)

        return result

    def _update_attrs(self, attrs, ct_id):
        current_model = ContentType.objects.get(id=ct_id).model_class()
        attrs["data-autocomplete-url"] = self.get_autocomplete_url(
            current_model,
        )
        attrs["data-app"] = current_model._meta.app_label
        attrs["data-model"] = current_model._meta.model_name
        attrs["data-change-url"] = self._get_change_url_template(current_model)

    def get_autocomplete_url(self, current_model):
        url_name = "admin:%s_%s_autocomplete"
        meta = current_model._meta

        return reverse(url_name % (meta.app_label, meta.model_name))

    def _get_change_url_template(self, current_model):
        meta = current_model._meta
        list_url = reverse(
            "admin:{0}_{1}_changelist".format(meta.app_label, meta.model_name),
        )
        return "{0}{{id}}/change/".format(list_url)



from contextlib import suppress

from .cache_calls import CacheCallsProfiler
from .db_queries import DatabaseQueriesProfiler
from .requests import RequestsProfiler

with suppress(ImportError):
    from .jaeger import JaegerProfiler



from typing import Tuple

from django.contrib.contenttypes.fields import GenericForeignKey
from django.core.exceptions import FieldDoesNotExist
from django.urls import NoReverseMatch
from django.utils.html import format_html

from jnt_django_toolbox.admin.helpers.urls import admin_url_provider
from jnt_django_toolbox.helpers.objects import copy_func

EMPTY_VALUE = "-"


def add_object_link_present_function(model_admin, field_name):
    field = "_{0}_object_link".format(field_name)

    copied_func = copy_func(_get_display_related_field)
    copied_func.__defaults__ = (field_name,)

    setattr(model_admin, field, copied_func)
    getattr(model_admin, field).__dict__.update(
        {"short_description": _get_short_description(model_admin, field_name)},
    )
    return format_html(field)


def object_change_link(obj, empty_description="-", field_present=None) -> str:
    if not obj:
        return empty_description

    obj_present = str(getattr(obj, field_present) if field_present else obj)

    try:
        return format_html(
            '<a href="{0}">{1}</a>',
            admin_url_provider.change_url(obj),
            obj_present,
        )
    except NoReverseMatch:
        return "{0} (id: {1})".format(obj, obj.id)


def _parse_list_field_attr(attr) -> Tuple[str, str]:
    unpack_attr = attr.split("__")
    return unpack_attr[0], unpack_attr[1] if len(unpack_attr) > 1 else None


def get_display_for_gfk(obj) -> str:
    if not obj:
        return None

    return format_html(
        '<span class="gfk-object-type">{0}</span> {1}'.format(
            obj._meta.model_name,
            object_change_link(obj),
        ),
    )


def _get_display_related_field(target_obj, attr):
    attr, field_name = _parse_list_field_attr(attr)

    obj = getattr(target_obj, attr, None)

    if not obj:
        return format_html(EMPTY_VALUE)

    model_field = target_obj._meta.get_field(attr)

    if model_field.many_to_many or model_field.one_to_many:
        return get_display_for_many(obj.all(), field_name)
    elif isinstance(model_field, GenericForeignKey):
        return get_display_for_gfk(obj)

    return object_change_link(obj)


def get_display_for_many(objects, field_present=None) -> str:
    if not objects.exists():
        return EMPTY_VALUE

    return format_html(
        ", ".join(
            map(
                object_change_link,
                (
                    getattr(instance, field_present)
                    if field_present
                    else instance
                    for instance in objects
                ),
            ),
        ),
    )


def _get_short_description(model_admin, field_name) -> str:
    """Get verbose name for target field."""
    default_present = field_name.split("__")[-1].replace("_", " ")
    try:
        field = model_admin.model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return default_present

    return getattr(field, "verbose_name", default_present)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jnt_django_toolbox',
 'jnt_django_toolbox.admin',
 'jnt_django_toolbox.admin.fields',
 'jnt_django_toolbox.admin.filters',
 'jnt_django_toolbox.admin.helpers',
 'jnt_django_toolbox.admin.mixins',
 'jnt_django_toolbox.admin.views',
 'jnt_django_toolbox.consts',
 'jnt_django_toolbox.context_managers',
 'jnt_django_toolbox.decorators',
 'jnt_django_toolbox.forms',
 'jnt_django_toolbox.forms.fields',
 'jnt_django_toolbox.forms.widgets',
 'jnt_django_toolbox.forms.widgets.readonly',
 'jnt_django_toolbox.helpers',
 'jnt_django_toolbox.management',
 'jnt_django_toolbox.management.commands',
 'jnt_django_toolbox.models',
 'jnt_django_toolbox.models.fields',
 'jnt_django_toolbox.profiling',
 'jnt_django_toolbox.profiling.db',
 'jnt_django_toolbox.profiling.decorators',
 'jnt_django_toolbox.profiling.profilers',
 'jnt_django_toolbox.templatetags']

package_data = \
{'': ['*'],
 'jnt_django_toolbox': ['static/jnt_django_toolbox/css/autocomplete_filter/*',
                        'static/jnt_django_toolbox/css/widgets/*',
                        'static/jnt_django_toolbox/images/*',
                        'static/jnt_django_toolbox/js/autocomplete_filter/*',
                        'static/jnt_django_toolbox/js/widgets/*',
                        'templates/jnt_django_toolbox/admin/*',
                        'templates/jnt_django_toolbox/autocomplete_filter/*',
                        'templates/jnt_django_toolbox/widgets/*']}

install_requires = \
['dateparser', 'django>=4.1']

extras_require = \
{'jaeger': ['jaeger-client']}

setup_kwargs = {
    'name': 'jnt-django-toolbox',
    'version': '0.8.0',
    'description': '',
    'long_description': 'None',
    'author': 'junte',
    'author_email': 'tech@junte.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from django.apps import apps
from django.contrib.contenttypes.fields import (
    GenericForeignKey as BaseGenericForeignKey,
)

from jnt_django_toolbox.admin.helpers.urls import admin_url_provider


class GenericForeignKey(BaseGenericForeignKey):
    def __init__(self, *args, **kwargs):
        self.related_models = kwargs.pop("related_models", None)
        super().__init__(*args, **kwargs)

    def get_related_models(self):  # noqa: WPS615
        if self.related_models is None:
            self.related_models = [
                model
                for model in apps.get_models()
                if self.has_autocomplete_url(model)
            ]
        elif callable(self.related_models):
            self.related_models = list(self.related_models())
        return [
            apps.get_model(*m.split(".")) if isinstance(m, str) else m
            for m in self.related_models
        ]

    def has_autocomplete_url(self, model):
        return bool(admin_url_provider.autocomplete_url(model))



from typing import Any, Dict, List

from django import forms
from django.apps import apps

DEFAULT_PERMISSIONS = ("add", "change", "delete", "view")


class PermissionSelectMultipleWidget(forms.CheckboxSelectMultiple):
    """
    A Widget for displaying all permissions in system.

    forms.py::

        from jnt_admin_tools.widgets import PermissionSelectMultipleWidget
        from django import forms
        from django.contrib.auth.models import Group

        class GroupAdminForm(forms.ModelForm):
            class Meta:
                model = Group
                fields = '__all__'
                widgets = {
                    'permissions': PermissionSelectMultipleWidget
                }

    admin.py::

        from django.contrib import admin
        from django.contrib.auth.models import Group
        from test_app.forms import GroupAdminForm

        @admin.register(Group)
        class GroupAdmin(admin.ModelAdmin):
            form = GroupAdminForm

    .. image:: images/widgets/permission_select_multiple_widget.png
    """

    class Media:
        css = {
            "all": ("jnt_django_toolbox/css/widgets/permissions.css",),
        }

    template_name = "jnt_django_toolbox/widgets/permissions.html"
    custom_permission_types: List[str] = []
    groups_permissions: List[str] = []
    extra_data: Dict[str, Any] = {}

    def get_context(self, name, value, attrs):
        if value is None:
            value = []

        context = {
            "name": name,
            "value": value,
            "table": self.get_table(),
            "groups_permissions": self.groups_permissions,
            "default_permission_types": DEFAULT_PERMISSIONS,
            "custom_permission_types": self.custom_permission_types,
            "is_readonly": False,
        }
        context.update(self.extra_data)

        return context

    def get_table(self):
        table = []
        row = None
        last_app = None
        last_model = None

        for permission in self.choices.queryset.select_related(
            "content_type",
        ).all():
            model_part = "_{0}".format(permission.content_type.model)
            permission_type = permission.codename
            if permission_type.endswith(model_part):
                permission_type = permission_type[: -len(model_part)]

            model_class = permission.content_type.model_class()

            app_config = (
                model_class._meta.app_config
                if model_class
                else apps.get_app_config(permission.content_type.app_label)
            )

            model_verbose_name = (
                model_class._meta.verbose_name if model_class else None
            )
            model_class_name = (
                model_class._meta.model_name if model_class else None
            )

            all_permissions = list(DEFAULT_PERMISSIONS) + list(
                self.custom_permission_types,
            )

            if permission_type not in all_permissions:
                self.custom_permission_types.append(permission_type)

            is_app_or_model_different = (
                last_model != model_class or last_app != app_config
            )
            if is_app_or_model_different:
                row = {
                    "model": model_verbose_name,
                    "model_class": model_class,
                    "model_class_name": model_class_name,
                    "app_config": app_config,
                    "app_label": permission.content_type.app_label,
                    "permissions": {},
                }

            row["permissions"][permission_type] = permission

            if is_app_or_model_different:
                table.append(row)

            last_app = app_config
            last_model = model_class

        return table



import types

from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.admin.widgets import SELECT2_TRANSLATIONS, get_language
from django.db.models.fields.related_descriptors import ManyToManyDescriptor
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from jnt_django_toolbox.forms.widgets import (
    AutocompleteSelect,
    AutocompleteSelectMultiple,
)

REPLACE_SYMBOLS = types.MappingProxyType(
    {
        "&quot;": '"',
        "&lt;": "<",
        "&gt;": ">",
    }
)


class AutocompleteFilter(admin.SimpleListFilter):
    template = (
        "jnt_django_toolbox/autocomplete_filter/autocomplete_filter.html"
    )
    title = ""
    field_name = ""
    is_placeholder_title = False
    widget_attrs = {}
    rel_model = None
    is_multiple = False

    class Media:
        extra = "" if settings.DEBUG else ".min"
        i18n_name = SELECT2_TRANSLATIONS.get(get_language())
        i18n_file = (
            ("admin/js/vendor/select2/i18n/%s.js" % i18n_name,)
            if i18n_name
            else ()
        )
        js = (
            (
                "admin/js/vendor/jquery/jquery%s.js" % extra,
                "admin/js/vendor/select2/select2.full%s.js" % extra,
            )
            + i18n_file
            + (
                "admin/js/jquery.init.js",
                "admin/js/autocomplete.js",
                "jnt_django_toolbox/js/autocomplete_filter/autocomplete-filter.js",
            )
        )
        css = {
            "screen": (
                "admin/css/vendor/select2/select2%s.css" % extra,
                "admin/css/autocomplete.css",
                "jnt_django_toolbox/css/autocomplete_filter/autocomplete-fix.css",
            ),
        }

    def __init__(self, request, params, model, model_admin):
        if self.parameter_name:
            raise AttributeError(
                "Rename attribute 'parameter_name' to "
                "'field_name' for {0}".format(self.__class__)
            )
        self.parameter_name = "{0}__id__{1}".format(
            self.field_name,
            "in" if self.is_multiple else "exact",
        )
        super().__init__(request, params, model, model_admin)

        if self.rel_model:
            model = self.rel_model

        remote_field = self.get_remote_field(model)
        formfield = self.get_formfield(model, remote_field, model_admin)

        attrs = self.widget_attrs.copy()
        attrs["id"] = "id-{0}-changelist-filter".format(self.field_name)

        if self.is_placeholder_title:
            attrs["data-placeholder"] = _("MSG_BY {title}").format(
                title=self.title,
            )

        self.rendered_widget = self.get_rendered_widget(formfield, attrs)

    @property
    def media(self):
        """Collect media files."""
        return forms.Media(js=self.Media.js, css=self.Media.css)

    def has_output(self):
        return True

    def lookups(self, request, model_admin):
        return ()

    def value(self):
        parameter_value = self.used_parameters.get(self.parameter_name, "")
        if not parameter_value:
            return ""

        if self.is_multiple:
            parameter_value = parameter_value.split(",")

        return parameter_value

    def queryset(self, request, queryset):
        if self.value():
            queryset = queryset.filter(
                **{self.parameter_name: self.value()}
            ).distinct()
        return queryset

    def get_field_queryset(self, model):
        field = getattr(model, self.field_name)
        if isinstance(field, ManyToManyDescriptor):
            return field.rel.model.objects.all()

        return field.get_queryset()

    def get_formfield(self, model, remote_field, model_admin):
        queryset = self.get_field_queryset(model)

        if self.is_multiple:
            return forms.ModelMultipleChoiceField(
                widget=AutocompleteSelectMultiple(
                    remote_field,
                    model_admin.admin_site,
                ),
                queryset=queryset,
                required=False,
            )

        return forms.ModelChoiceField(
            widget=AutocompleteSelect(remote_field, model_admin.admin_site),
            queryset=queryset,
            required=False,
        )

    def get_remote_field(self, model):
        return getattr(model, self.field_name).field

    def get_rendered_widget(self, field, attrs):  # noqa: WPS615
        rendered_widget = field.widget.render(
            name=self.parameter_name,
            value=self.value(),
            attrs=attrs,
        )

        if self.is_multiple:
            for old_symbol, new_symbol in REPLACE_SYMBOLS.items():
                rendered_widget = rendered_widget.replace(
                    old_symbol,
                    new_symbol,
                )
            rendered_widget = format_html(rendered_widget)

        return rendered_widget



from functools import partialmethod

from django import forms
from django.contrib.contenttypes.models import ContentType
from django.template.response import TemplateResponse

from jnt_django_toolbox.forms.widgets import GenericForeignKeyWidget
from jnt_django_toolbox.models.fields import GenericForeignKey


class GenericForeignKeyAdminMixin:
    def get_form(self, request, obj=None, change=False, **kwargs):
        form = super().get_form(request, obj=obj, change=change, **kwargs)
        self._update_gr_fields(form, obj)
        return form

    def get_fields(self, *args, **kwargs):
        fields = super().get_fields(*args, **kwargs)
        return self._update_fields(fields)

    def get_fieldsets(self, *args, **kwargs):
        fieldsets = super().get_fieldsets(*args, **kwargs)
        for _, fieldset_fields in fieldsets:
            fieldset_fields["fields"] = self._update_fields(
                fieldset_fields["fields"]
            )
        return fieldsets

    def changeform_view(self, *args, **kwargs):
        changeform_view = super().changeform_view(*args, **kwargs)

        self._update_gfk_in_inlines(changeform_view)
        return changeform_view

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        form_field = super().formfield_for_foreignkey(
            db_field,
            request,
            **kwargs,
        )
        generic_relation_fields = self._get_generic_relation_fields(self.model)
        current_generic_field = None
        for generic_field in generic_relation_fields:
            if generic_field.ct_field == db_field.name:
                current_generic_field = generic_field
                break

        if not current_generic_field:
            return form_field

        is_required = form_field.widget.is_required
        form_field.widget = GenericForeignKeyWidget(
            db_field.remote_field,
            self.admin_site,
            using=kwargs.get("using"),
            attrs=self._get_content_type_autocomplete_attrs(
                current_generic_field
            ),
        )
        form_field.widget.is_required = is_required

        return form_field

    def _update_fields(self, fields):
        generic_relation_fields = self._get_generic_relation_fields(self.model)
        if not generic_relation_fields:
            return fields

        fields = list(fields)

        for generic_field in generic_relation_fields:
            if generic_field.name not in fields:
                continue

            if generic_field.ct_field not in fields:
                fields[
                    fields.index(generic_field.name)
                ] = generic_field.ct_field

            if generic_field.fk_field not in fields:
                fields.append(generic_field.fk_field)

            if generic_field.name in fields:
                fields.pop(fields.index(generic_field.name))

        return tuple(fields)

    def _update_gr_fields(self, form, instance) -> None:
        if not instance:
            return

        generic_relation_fields = self._get_generic_relation_fields(self.model)

        for field in generic_relation_fields:
            if (
                field.ct_field in form.base_fields
                and field.fk_field in form.base_fields
            ):
                fk_present = getattr(instance, field.name)
                fk_present = str(fk_present) if fk_present else ""
                form.base_fields[field.fk_field].widget = forms.HiddenInput(
                    attrs={
                        "data-present": fk_present,
                    },
                )
                clean_func = partialmethod(
                    self._clean_fk_field,
                    fk_field=field.fk_field,
                    ct_field=field.ct_field,
                )
                setattr(form, "clean_{0}".format(field.fk_field), clean_func)

    def _get_generic_relation_fields(self, model):
        return [
            field
            for field in model._meta.get_fields()
            if self._is_generic_relation_field(field)
        ]

    def _is_generic_relation_field(self, field) -> bool:
        return (
            not field.auto_created
            and field.is_relation
            and isinstance(field, GenericForeignKey)
        )

    def _get_content_type_autocomplete_attrs(self, generic_field):
        return {
            "class": "generic-foreign-key-field",
            "data-ct--field": generic_field.ct_field,
            "data-fk--field": generic_field.fk_field,
            "data-gf--name": generic_field.name,
            "data-gfk--models": ",".join(
                str(model_id)
                for model_id in self._get_gfk_ids_related_models(generic_field)
            ),
        }

    def _get_gfk_ids_related_models(self, gfk):
        related_models = gfk.get_related_models()

        content_types = ContentType.objects.get_for_models(*related_models)

        return [x.id for x in content_types.values()]

    def _update_gfk_in_inlines(self, changeform_view) -> None:
        if not isinstance(changeform_view, TemplateResponse):
            return
        inline_formsets = changeform_view.context_data.get(
            "inline_admin_formsets"
        )
        if not inline_formsets:
            return

        for inline_formset in inline_formsets:
            generic_fields = self._get_generic_relation_fields(
                inline_formset.opts.model
            )
            if not generic_fields:
                continue

            for inline_form in inline_formset.forms:
                instance = inline_form.instance

                for generic_field in generic_fields:
                    if generic_field.fk_field in inline_form.fields:
                        hidden_widget = inline_form.fields[
                            generic_field.fk_field
                        ].widget
                        generic_value = getattr(
                            instance, generic_field.name, ""
                        )
                        hidden_widget.attrs["data-present"] = str(
                            generic_value
                        )

    def _clean_fk_field(self, current_form, **kwargs):
        fk_value = current_form.cleaned_data[kwargs["fk_field"]]
        ct_value = current_form.cleaned_data[kwargs["ct_field"]]

        if ct_value and not fk_value:
            current_form.add_error(
                kwargs["ct_field"],
                forms.Field.default_error_messages["required"],
            )
        return fk_value



import typing as ty
from dataclasses import dataclass

from django import forms
from django.conf import settings
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.utils.translation import gettext_lazy as _
from django.views.generic import FormView


class _EmptyForm(forms.Form):
    """Empty form."""


@dataclass
class AdminPageBreadcrumb:
    title: str
    href: str | None = None


class BaseAdminPageView(PermissionRequiredMixin, FormView):
    site_title = _("Django site admin")
    site_header = _("Django administration")
    site_url = "/"
    title = "Admin Page"
    help_text: str | None = None
    template_name = "jnt_django_toolbox/admin/admin_page.html"
    submit_label = ""
    success_url = "."
    message_safe = True
    form_class = _EmptyForm
    permission_required: str | ty.Iterable[str] | None = None

    @property
    def media(self):
        extra = "" if settings.DEBUG else ".min"
        js = [
            "vendor/jquery/jquery{0}.js".format(extra),
            "jquery.init.js",
        ]
        return forms.Media(
            js=["admin/js/{0}".format(url) for url in js],
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            site_url=self.site_url,
            site_title=self.site_title,
            site_header=self.site_header,
            title=self.title,
            help_text=self.help_text,
            media=self.media + context["form"].media,
            message_safe=self.message_safe,
            request=self.request,
            user=self.request.user,
            has_permission=self.has_permission(),
            submit_label=self.submit_label,
            breadcrumbs=self.get_breadcrumbs(),
        )
        return context

    def has_permission(self) -> bool:
        """Return True if the given HttpRequest has permission to view page."""
        user = self.request.user
        if not user.is_authenticated:
            return False

        if not (user.is_active and user.is_staff):
            return False

        perms = self.get_permission_required()
        if perms:
            return self.request.user.has_perms(perms)

        return True

    def get_permission_required(self) -> ty.Iterable[str]:  # noqa: WPS615
        if not self.permission_required:
            return ()
        elif isinstance(self.permission_required, str):
            return (self.permission_required,)

        return self.permission_required

    def get_admin_permissions(self) -> list[str]:
        return []

    def get_breadcrumbs(self) -> list[AdminPageBreadcrumb]:
        return [
            AdminPageBreadcrumb(
                title=self.title,
            ),
        ]



from django.contrib.admin.utils import flatten_fieldsets
from django.db.models import ForeignKey, ManyToManyField

from jnt_django_toolbox.admin.helpers.widgets import render_autocomplete_badge
from jnt_django_toolbox.forms.widgets import autocomplete_select


class AutocompleteWidgetsUpdateAdminMixin:
    """Update autocomplete widgets for models."""

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        """
        Get a form Field for a ForeignKey.

        Override only autocomplete widget.
        """
        if "widget" not in kwargs:
            if db_field.name in self.get_autocomplete_fields(request):
                kwargs["widget"] = autocomplete_select.AutocompleteSelect(
                    db_field,
                    self.admin_site,
                    using=kwargs.get("using"),
                )

        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def formfield_for_manytomany(self, db_field, request, **kwargs):
        """
        Get a form Field for a ManyToManyField.

        Override only autocomplete widget.
        """
        if not db_field.remote_field.through._meta.auto_created:
            return None

        if "widget" not in kwargs:
            autocomplete_fields = self.get_autocomplete_fields(request)
            if db_field.name in autocomplete_fields:
                kwargs[
                    "widget"
                ] = autocomplete_select.AutocompleteSelectMultiple(
                    db_field,
                    self.admin_site,
                    using=kwargs.get("using"),
                )

        return super().formfield_for_manytomany(db_field, request, **kwargs)


class AutocompleteFieldsAdminMixin(AutocompleteWidgetsUpdateAdminMixin):
    def get_autocomplete_fields(self, request):
        autocomplete_fields = super().get_autocomplete_fields(request)

        if autocomplete_fields:
            return autocomplete_fields

        # we returns all model related fields
        relation_fields = self._get_relation_fields()

        return tuple(relation_fields)

    def autocomplete_queryset(self, request, queryset):
        if not queryset.ordered:
            queryset = queryset.order_by(*self.get_ordering(request))

        return queryset

    def autocomplete_item_data(self, instance, request):
        """Get present for autocomplete item."""
        return {
            "id": instance.pk
            if isinstance(instance.pk, int)
            else str(instance.pk),
            "text": str(instance),
            "__badge__": render_autocomplete_badge(instance),
        }

    def _get_relation_fields(self):
        return (
            field.name
            for field in self.model._meta.get_fields()
            if self._is_relation_field(field)
        )

    def _get_admin_fields(self, request):
        if self.fields:
            return self.get_fields(request)
        elif self.fieldsets:
            return flatten_fieldsets(self.get_fieldsets(request))

    def _is_relation_field(self, field):
        return isinstance(field, (ForeignKey, ManyToManyField)) and bool(
            self.admin_site._registry.get(field.remote_field.model)
        )
